/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

// PicoWidget : Scrollbar
// Copyrights (c) PicoPeta Simputers Pvt. Ltd.
// Programmer: Santhosh N

#include "scrollbar.h"
#include "image.h"
#include "widget_factory.h"
#include <X11/extensions/shape.h>
#include <stdlib.h>
#include <string.h>
#include "theme.h"
#include "eventloop.h"
#include "display.h"

#define SCROLL_DELAY 130000

typedef enum _scrollbar_states 
{
	SCROLL_NONE, SCROLL_UP, SCROLL_DOWN, SCROLL_SLIDER, SCROLL_TROUGH
} scrollbar_states;

static scrollbar_states sbar_state;
static px_timer *timer=NULL;

px_widget* scrollbar_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();
	px_theme* theme = theme_get_theme();

	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0,
							CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);

	// Initialize common widget variables
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = 16;
	widget->wtype = SCROLLBAR;
	widget->child_of = NULL;
	widget->use_pixmap = True;
	if(tag != NULL) {
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}else
		widget->tag = NULL;

	widget->background.background_pixmap = theme_get_pixmap(theme, SCROLLBAR_TROUGH);

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);
	
	return widget;
}

void scrollbar_set_parameters(px_widget* widget)
{
	widget->default_callbacks[BUTTON_PRESSED] = scrollbar_default_button_press;
	widget->default_callbacks[BUTTON_RELEASED] = scrollbar_default_button_release;
	widget->default_callbacks[MOUSE_LEAVE] = scrollbar_default_button_release;
	widget->default_callbacks[MOTION] = scrollbar_default_button_motion;

	widget->redraw = scrollbar_redraw;
	widget->destroy = scrollbar_destroy;

	widget->specific.scrollbar.slider_start = 0.0;
	widget->specific.scrollbar.slider_length = 1.0;
	widget->specific.scrollbar.increment = 0.1;
	widget->specific.scrollbar.x = 0;
	widget->specific.scrollbar.y = 0;

	// Initalizing widget specific callbacks
	widget->specific.scrollbar.set_slider_position = scrollbar_set_slider_position;
	widget->specific.scrollbar.set_slider_length = scrollbar_set_slider_length;
	widget->specific.scrollbar.set_increment = scrollbar_set_increment;
	widget->specific.scrollbar.redraw_slider = scrollbar_redraw_slider;
	widget->specific.scrollbar.redraw_top_button = scrollbar_redraw_top_button;
	widget->specific.scrollbar.redraw_bottom_button = scrollbar_redraw_bottom_button;
	widget->specific.scrollbar.intimate_viewport_changed = NULL;
	widget->specific.scrollbar.get_slider_position = scrollbar_get_slider_position;
	widget->specific.scrollbar.get_increment = scrollbar_get_increment;
}

Bool scrollbar_handle_up_arrow(px_widget* widget, px_event* event)
{
	px_scrollbar* scrollbar = &(widget->specific.scrollbar);
	float nstart;

	nstart = scrollbar->slider_start - scrollbar->increment;

	if(nstart < 0.0) {
		nstart = 0.0;
		if(timer != NULL) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			eventloop->delete_timer(eventloop, timer);
			timer = NULL;
		}
	}

	scrollbar->slider_start = nstart;
	scrollbar->redraw_slider(widget);
		
	if(scrollbar->intimate_viewport_changed != NULL)
		scrollbar->intimate_viewport_changed (widget->child_of, scrollbar->slider_start);

	return True;
}

Bool scrollbar_handle_down_arrow(px_widget* widget, px_event* event)
{
	px_scrollbar* scrollbar = &(widget->specific.scrollbar);
	float nstart;

	nstart = scrollbar->slider_start + scrollbar->increment;

	if(nstart > 1.0) {
		nstart = 1.0 ;
		if(timer != NULL) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			eventloop->delete_timer(eventloop, timer);
			timer = NULL;
		}
	}

	scrollbar->slider_start = nstart;
	scrollbar->redraw_slider(widget);
		
	if(scrollbar->intimate_viewport_changed != NULL)
		scrollbar->intimate_viewport_changed (widget->child_of, scrollbar->slider_start);

	return True;
}
	
Bool scrollbar_default_button_press(px_widget* widget, px_event* event)
{
	px_scrollbar* scrollbar = &(widget->specific.scrollbar);
	int slider_start, slider_length;
		px_eventloop* eventloop = eventloop_get_eventloop();

	slider_length = (int)(scrollbar->slider_length*(widget->height-32));
	if(slider_length<24)
		slider_length = 24;
	slider_start = (int)(scrollbar->slider_start*(widget->height-32-slider_length));

	if(event->y >= 0 && event->y <= 16) {	// Up arrow
		sbar_state = SCROLL_UP;
		if(scrollbar->slider_start > 0.0) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			scrollbar_handle_up_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, SCROLL_DELAY, scrollbar_handle_up_arrow, widget, event, True);
		}
		scrollbar->redraw_top_button(widget);
	} else if(event->y >= (widget->height-16) && event->y <= (widget->height-1)) {	// Down arrow
		sbar_state = SCROLL_DOWN;
		if(widget->height-16 > slider_start+slider_length+16) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			scrollbar_handle_down_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, SCROLL_DELAY, scrollbar_handle_down_arrow, widget, event, True);
		}
		scrollbar->redraw_bottom_button(widget);
	} else if(event->y > 11 && event->y < slider_start+16) {	// Trough above slider
		sbar_state = SCROLL_TROUGH;
		if(scrollbar->slider_start > 0) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			scrollbar_handle_up_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, SCROLL_DELAY, scrollbar_handle_up_arrow, widget, event, True);
		}
	} else if(event->y > slider_start+slider_length+16 && event->y < widget->height-16) { // Trough below slider
		sbar_state = SCROLL_TROUGH;
		if(slider_start+slider_length+16 < widget->height-16) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			scrollbar_handle_down_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, SCROLL_DELAY, scrollbar_handle_down_arrow, widget, event, True);
		}
	} else {	// Slider
		sbar_state = SCROLL_SLIDER;
		scrollbar->x = event->x;
		scrollbar->y = event->y;
	}
		
	eventloop->grab(eventloop, widget);

	return True;
}
		
Bool scrollbar_default_button_release(px_widget* widget, px_event* event)
{
	px_eventloop* eventloop = eventloop_get_eventloop();
	px_scrollbar* scrollbar = &(widget->specific.scrollbar);

	eventloop->grab(eventloop, NULL);
	
	if(timer != NULL) {
		eventloop->delete_timer(eventloop, timer);
		timer = NULL;
	}

	sbar_state = SCROLL_NONE;
	scrollbar->x = 0;
	scrollbar->y = 0;

	widget->redraw(widget);
	widget->show(widget);

	return True;
}

Bool scrollbar_default_button_motion(px_widget* widget, px_event* event)
{
	px_scrollbar* scrollbar = &(widget->specific.scrollbar);
	int fg;
	XEvent xev, xev_copy;

	// Check if within Slider
	if(sbar_state == SCROLL_SLIDER && event->y >= 16 && event->y <= (widget->height-16)) {
		float nstart;
		int slider_length;

		slider_length = (int)(scrollbar->slider_length*(widget->height-32));
		if(slider_length<24)
			slider_length = 24;
		nstart = (scrollbar->slider_start*(widget->height-32-slider_length)+(event->y-scrollbar->y))/(widget->height-32-slider_length);

		// Make sure that slider is within the designated area
		if(nstart<0.0)
			nstart = 0.0;
		else if(nstart>1.0)
			nstart = 1.0;

		scrollbar->slider_start = nstart;
		scrollbar->redraw_slider(widget);
		
		if(scrollbar->intimate_viewport_changed != NULL)
			scrollbar->intimate_viewport_changed(widget->child_of, scrollbar->slider_start);

		scrollbar->x = event->x;
		scrollbar->y = event->y;
	}

	// Bypass extra motion events
	fg = 0;
	while(XPending(display_get_Xdisplay())) {
		XNextEvent(display_get_Xdisplay(), &xev);
		if(!fg && xev.type==MotionNotify) {
			fg = 1;
			xev_copy = xev;
		}
		if(xev.type==ButtonRelease) {
			fg = 1;
			xev_copy = xev;
			break;
		}
	}
	if(fg)
		XPutBackEvent(display_get_Xdisplay(), &xev_copy);

	return True;
}

void scrollbar_redraw_top_button(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
	GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	px_pixmap* show_pixmap = NULL;
	px_theme* theme = theme_get_theme();

	if(sbar_state == SCROLL_UP)
		show_pixmap = theme_get_pixmap(theme, SCROLLBAR_UP_ACTIVE);
	else
		show_pixmap = theme_get_pixmap(theme, SCROLLBAR_UP);
	
	XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc, 
				0, 0, 16, 16, 0, 0); 
	
	XClearArea(display, widget->win, 0, 0, 16, 16, False);

	XFreeGC(display, gc);
}

void scrollbar_redraw_bottom_button(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
	GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	px_pixmap* show_pixmap = NULL;
	px_theme* theme = theme_get_theme();
	
	if(sbar_state == SCROLL_DOWN)
		show_pixmap = theme_get_pixmap(theme, SCROLLBAR_DOWN_ACTIVE);
	else
		show_pixmap = theme_get_pixmap(theme, SCROLLBAR_DOWN);
	
	XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc, 
				0, 0, 16, 16, 0, widget->height - 16); 

	XClearArea(display, widget->win,  0, widget->height-16, 16, 16, False);
	
	XFreeGC(display, gc);
}
	
void scrollbar_redraw_slider(px_widget* widget)
{
	px_scrollbar* scrollbar = &(widget->specific.scrollbar);
	Display* display = display_get_Xdisplay();
	px_theme* theme = theme_get_theme();
	px_pixmap *show_pixmap = NULL;
	GC gc, gcxor, gcand;
	int slider_start, slider_length;

	slider_length = (int)(scrollbar->slider_length*(widget->height-32));
	if(slider_length<24) 
		slider_length = 24;
	slider_start = (int)(scrollbar->slider_start*(widget->height-32-slider_length));

	if(widget->use_pixmap && (slider_start+slider_length)<widget->height) {
		// Create required GCs
		gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);

		gcxor = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetForeground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
		XSetBackground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
		XSetFunction(display, gcxor, GXxor);
		XSetFillStyle(display, gcxor, FillSolid);
		XSetGraphicsExposures(display, gcxor, False);

		gcand = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetForeground(display, gcand, 0);
		XSetBackground(display, gcand, ~0);
		XSetFunction(display, gcand, GXand);
		XSetFillStyle(display, gcand, FillSolid);
		XSetGraphicsExposures(display, gcand, False);

		// Draw the scrollbar pixmaps
		show_pixmap = theme_get_pixmap(theme, SCROLLBAR_TROUGH2);
		XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
							0, 0, 16, widget->height-32, 0, 16);
		
		show_pixmap = theme_get_pixmap(theme, SCROLLBAR_SLIDER);
		if(show_pixmap->mask!=None) {		// Transperant Slider
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 16, slider_length/2+1, 0, slider_start+16);
			XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand,
								0, 0, 16, slider_length/2+1, 0, slider_start+16, 1L);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 16, slider_length/2+1, 0, slider_start+16);

			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 320-slider_length/2, 16, slider_length/2, 0, slider_start+16+slider_length/2);
			XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand,
								0, 320-slider_length/2, 16, slider_length/2, 0, slider_start+16+slider_length/2, 1L);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 320-slider_length/2, 16, slider_length/2, 0, slider_start+16+slider_length/2);
		} else {
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
								0, 0, 16, slider_length/2+1, 0, slider_start+16);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
								0, 320-slider_length/2, 16, slider_length/2, 0, slider_start+16+slider_length/2);
		}

		show_pixmap = theme_get_pixmap(theme, SCROLLBAR_NOTCH);
		if(show_pixmap->mask!=None) {		// Transperant Slider Notch
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 10, 10, 3, (2*slider_start+slider_length)/2-5+16);
			XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand,
								0, 0, 10, 10, 3, (2*slider_start+slider_length)/2-5+16, 1L);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 10, 10, 3, (2*slider_start+slider_length)/2-5+16);
		} else {
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
								0, 0, 10, 10, 3, (2*slider_start+slider_length)/2-5+16);
		}
		
		XClearArea(display, widget->win, 0, 16, widget->width, widget->height-32, False);

		XFreeGC(display, gcand);
		XFreeGC(display, gcxor);
		XFreeGC(display, gc);
	}
}

void scrollbar_redraw(px_widget* widget)
{
	// Draw the arrows
	scrollbar_redraw_top_button(widget);
	scrollbar_redraw_bottom_button(widget);

	// Draw the slider
	scrollbar_redraw_slider(widget);

	widget_set_event_mask(widget, widget->emask);
}

void scrollbar_destroy(px_widget* widget)
{
	widget_destroy(widget);
}

void scrollbar_set_slider_position(px_widget* widget, float s)
{
	if(s<0.0)
		widget->specific.scrollbar.slider_start = 0.0;
	else if(s>1.0)
		widget->specific.scrollbar.slider_start = 1.0;
	else
		widget->specific.scrollbar.slider_start = s;
}

void scrollbar_set_slider_length(px_widget* widget, float s)
{
	if(s<0.0)
		widget->specific.scrollbar.slider_length = 0.0;
	else if(s>1.0)
		widget->specific.scrollbar.slider_length = 1.0;
	else
		widget->specific.scrollbar.slider_length = s;
}

void scrollbar_set_increment(px_widget* widget, float i)
{
	if(i>=0.0 && i<=1.0)
		widget->specific.scrollbar.increment = i;
}

float scrollbar_get_slider_position(px_widget* widget)
{
	return widget->specific.scrollbar.slider_start;
}

float scrollbar_get_increment(px_widget* widget)
{
	return widget->specific.scrollbar.increment;
}
